#!/usr/bin/env python3

print("""
Dieses Python3-Skript erstellt Höhenmodelle für CAD und 3D-Druck.

Mit zwei Dezimalgradkoordinaten geben Sie die Eckpunkte einer rechteckigen
Geländeoberfläche an, für die dieses Programm mehrere 3D-Dateien erzeugt:

 - ein Höhenmodell im STL-Format für 3D-Drucker,
 - eine DXF-Datei mit 3D-Flächen, die in Regionen umgewandelt und zu einem
   Volumenmodell geheftet werden können,
 - eine auf die von Ihnen gewählte Auswahl und Auflösung reduzierte XYZ-Datei, 
 - eine importfähige DGM-Datei im TXT-Format für BricsCADs TIN-Befehl,
 - ein Ortophoto des gewählten Bereichs (100 Pixel pro Höhenpunkt), sowie
 - eine Höhenkarte im PNG-Format, falls das Python-Modul Matplotlib
   installiert ist.
""")

# Das aus Drei- und Vierecken zusammengesetzte Oberflächenmodell in der
# DXF-Datei lässt sich durch Anwenden der beiden Befehle "REGION" und
# "DMHEFTEN" (BricsCAD) bzw. "FLÄCHEFORM" (AutoCAD) in ein Volumenmodell
# umwandeln.

# Das Programm wurde für die öffentlichen Höhendaten für digitale
# Geländemodelle in NRW geschrieben, lässt sich aber mit entsprechenden
# Anpassungen auch für andere DGM-Daten nutzen.

# Autor: Dipl.-Ing. Martin Vogel, Hochschule Bochum, 2017–2021

# Lizenz: Namensnennung - Weitergabe unter gleichen Bedingungen 3.0 Deutschland
# (CC BY-SA 3.0 DE) https://creativecommons.org/licenses/by-sa/3.0/de/

# Version 18.2 vom 1. Dezember 2021
# Orthophotos werden jetzt nur noch heruntergeladen, wenn osgeo/gdal
# vorhanden ist, da es sich um recht große Downloads handelt.

# Version 18.1 vom 1. Dezember 2021
# Anpassung für eine Änderung in der URL für die Orthophotos. Diese sind jetzt
# nicht mehr unter https://www.opengeodata.nrw.de/produkte/geobasis/lbi/dop/…
# …dop_jp2_f10/ abgelegt, sondern unter https://www.opengeodata.nrw.de/…
# …produkte/geobasis/lusat/dop/dop_jp2_f10/.

# Version 18 vom 13. April 2021
#   Anpassung an neues Dateinamenschema auf dem Geodatenserver.
#   "dgm1_32_280_5652_1_nw.xyz.gz" statt "dgm1_32280_5652_2_nw.xyz.gz"
#   Die Zonennummer ist nun vom Ostwert getrennt.
#   Die Kacheln sind nun nur noch 1×1 statt 2×2 km² groß und decken nun
#   dieselben Flächen wie die Orthophotos ab.
#   Bei der Umrechnung der Koordinaten entfällt der Korrekturwert 32000000.

# Version 17.2 vom 26. Juni 2020
#   Erzeugt jetzt beliebig große Orthophotos im JPG-Format aus den einzelnen
#   j2p-Kacheln.

# Version 17.1 vom 16. Juni 2020
#   Korrektur für die Eingabe kaputter Drei-Komma-Koordinaten aus Bing Maps.
#   Keine Anzeige des Matplotlib-Höhendiagramms mehr.

# Version 17 vom 11. Juni 2020
#   Herunterladen der Orthophotos des ausgewählten Rechtecks
#   Benötigt gdal zum Lesen der 4-Kanal-jp2-Dateien (RGBI mit Infrarotkanal).
#   Funktioniert momentan nur innerhalb einer Orthophotokachel.

# Version 16.1 vom 10. Juni 2020
#   Bugfix: Wenn das höchstgelegene Dreieck einer DXF-Datei exakt horizontal
#   lag, gab es einen Indexfehler bei der Farbberechnung in der Funktion hf.
#
#   Der Ordner für die DGM-Dateien ist jetzt fest auf "DGM1" eingestellt.
#
# Version 16.0 vom 29. Mai 2020
#   Die Download-URL der DGM-Dateien wurde schon wieder geändert.
#
#   Die alten, zu Bezirken zusammengefassten ZIP-Dateien sind nun in
#       https://www.opengeodata.nrw.de
#             /produkte/geobasis/hm/dgm1_xyz/dgm1_xyz_paketiert/
#   
#   Diese Riesenpakete werden vom Programm aber gar nicht mehr verwendet,
#   denn es gibt jetzt ein neues Verzeichnis mit einzelnen,
#   Gnuzip-komprimierten Dateien, die nur noch um die 20 MB groß sind und
#   viel besser geeignet sind, unter der URL
#       https://www.opengeodata.nrw.de
#             /produkte/geobasis/hm/dgm1_xyz/dgm1_xyz/
#   Das Programm liest diese Dateien nun direkt, ohne sie erst aufwendig
#   auspacken und zwischenspeichern zu müssen.

# Version 15.3 vom 22. Januar 2020
#   Die Download-URL der DGM-Dateien wurde geändert.
#   alt: https://www.opengeodata.nrw.de/produkte/geobasis/dgm/dgm1/
#   neu: https://www.opengeodata.nrw.de/produkte/geobasis/hm/dgm1_xyz/

# Version 15.2 vom 23. Juni 2019
#   Die Quadrate der Geländeoberfläche werden nun so in zwei Dreiecke
#   aufgeteilt, dass die Diagonale möglichst kurz ist. Die Oberfläche wirkt
#   dadurch, besonders bei geringen Auflösungen, deutlich realistischer.

# Version 15.1 vom 22. Juni 2019
#   3D-Flächen der DXF-Datei mit höhenabhängiger Indexfarbe.
#   Layer für alle Entities ist nun "Gelaendemodell.py".

# Version 15 vom 21. Juni 2019
#   Erzeugung von CAD-Scriptdateien aus dem Programm entfernt.
#   Seit der DXF-Unterstützung waren die SCR-Scripte eigentlich nur noch
#   eine überflüssige Spielerei.

# Version 14.3 vom 21. Juni 2019
#   In der DXF-Datei werden jetzt alle Entities explizit Layer "0" zugeordnet.
#   AutoCAD wirft sonst einen Importfehler aus.

# Version 14.2 vom 12. Juni 2019
#   Vor dem Herunterladen einer ZIP-Datei wird geprüft, ob sie nicht schon
#   existiert.
#   Es werden nur noch die gerade benötigten XYZ-Dateien aus der ZIP-Datei
#   ausgepackt.
#   Export des Matplotlib-Höhendiagramms als PNG statt PDF, weil der Export
#   von PDF- und SVG-Vektordaten hoher Auflösung extrem langsam ist.
#   Download jetzt mit Fortschrittsanzeige.

# Version 14.1 vom 12. Juni 2019
#   ZIP-Dateien mit Umlauten im Namen werden jetzt auch unter Microsoft Windows
#   heruntergeladen.

# Version 14 vom 9. Juni 2019
#   Fehlende Kacheln werden nun automatisch nachgeladen.
#   Eine Fortschrittsanzeige fehlt allerdings noch.

# Version 13.1 vom 6. Juni 2019
#   Bugfix, weil der Verzeichnisauswahldialog auf manchen Mac-OS-Geräten
#   nicht geschlossen wurde.
#   Alle alten C-Formatstrings durch f-Strings ersetzt. 

# Version 13 vom 10. Mai 2019
#   XYZ-Daten als TXT-Datei für Bricscad-BIM-Geländeoberflächen.

# Version 12 vom 3. Juni 2018
#   Anzeige und PDF-Export eines Höhendiagramms mittels Matplotlib.

# Version 11 vom 18. Juni 2017
#   Zu fehlenden Kacheln wird nun angegeben, in welchen Ortsdateien diese
#   zu finden sind. Dazu mussten alle 396 Dateien einmal heruntergeladen und
#   diese 269 Gigabyte ZIP-Dateien katalogisiert werden. -> Gelaendekatalog.csv

# Version 10 vom 15. Juni 2017
#   UTM-Koordinatenumrechnung jetzt ohne externes Modul und nach vereinfachtem
#   Algorithmus.

# Version 9 vom 13. Juni 2017
#   DXF-Export der Dreiecks- und Vierecksflächen aus Version 8.

# Version 8 vom 11. Juni 2017
#   Oberflächenmodell für CAD-Skript aus einzelnen Dreiecksflächen, wie in der
#   STL-Datei, Seitenflächen und Boden jedoch aus Vierecken.

# Version 7 vom 9. Juni 2017
#   Jetzt mit 3D-Netz, viel schneller als Volumenkörper im CAD-Programm.
#   Leider kann BricsCAD aus dieser Oberfläche keinen Volumenkörper heften.

# Version 6 vom 21. März 2017
#   Jetzt mit Log-Datei zur Garantie der Wiederholbarkeit.

# Version 5 vom 19. März 2017
#   Jetzt mit variabler Höhe der Unterseite. Es gibt in NRW tatsächlich Flächen,
#   die weit unterhalb des Meeresspiegels liegen. (Tagebau Etzweiler: -298,90m!)

# Version 4 vom 18. März 2017
#   Jetzt mit binärem STL-Export. Die Dateien sind nur noch halb so groß
#   und werden viel schneller gelesen als ASCII-STL-Dateien.

# Version 3 vom 17. März 2017
#   Einige mögliche Fehleingaben werden jetzt freundlich abgefangen.
#   Erweiterung um CAD-Skript für Dreiecksprismen, um den Minecraft-Effekt
#   loszuwerden.

# Version 2 vom 16. März 2017
#   Jetzt mit Eingabe der Höhenauflösung.
#   Außerdem werden nicht mehr alle gefundenen Punkte im Dictionary
#   abgelegt, sondern nur noch die später auch benötigten.

# Ideen für Erweiterungen:
#  - Grafische Oberfläche (mit Kontrolle der Kachelabdeckung) einbauen.
#  - In die DXF-Datei noch Werte für Zoom, Blickwinkel, Geokoordinaten
#    usw. eintragen.

import os
import sys
# Für die Anzeige des Downloadverzeichnisses
import webbrowser
# Für die binären Dateiformate
import struct
# Ohne Trigonometrie geht hier nichts.
from math import sqrt, sin, cos, tan, radians, degrees, floor
# Für den Download der Geodaten
import urllib.request
# Zum Auspacken der Daten
import gzip

# Das Zusammensetzen der Orthophotos wird mit osgeo/gdal realisiert:
try:
    from osgeo import gdal
except ModuleNotFoundError:
    gdal = None

# Das Programm soll auch ohne Matplotlib durchlaufen.
try:
    import matplotlib.pyplot as plt
except ModuleNotFoundError:
    plt = None
    

def sysexit():
    input("\nProgramm wird beendet. [Enter]")
    sys.exit()

def log(s, sichtbar=True):
    "Bildschirmmeldung mit Protokollierung"
    with open(basisname+".log","a") as logfile:
        logfile.write(f"{s}\n")
    if sichtbar:
        print(f"\n{s}")

def utm(Bg,Lg):
    "Umrechnung von Breitengrad und Längengrad in UTM-Koordinaten"
    # Quelle: http://www.ottmarlabonde.de/L1/UTMBeispielRechnung.htm
    # Literatur: A. Schödlbauer,
    #     Rechenformeln und Rechenbeispiele zur Landesvermessung,
    #     Teil 2, Herbert WichmannVerlag Karlsruhe

    L = radians(Lg)
    B = radians(Bg)
    tB = tan(B)
    cB = cos(B)

    # Halbachsen des WGS84-Ellipsoids:
    # a = 6378137.0
    # b = 6356752.314

    # Radiusreduzierung
    mH = 0.9996

    # c = a**2/b
    c = 6399593.626005325

    # eq = (a**2-b**2)/b**2
    eq = 0.006739496819936062

    # E0 = c*(1-3/4*eq+45/64*eq**2-175/256*eq**3+11025/16384*eq**4)
    E0 = 6367449.145759811

    # E2 = c*(-3/8*eq+15/32*eq**2-525/1024*eq**3+2205/4096*eq**4)
    E2 = -16038.508797800609

    # E4 = c*(15/256*eq**2-105/1024*eq**3+2205/16384*eq**4)
    E4 = 16.83262765753934
    
    # E6 = c*(-35/3072*eq**3+315/12288*eq**4)
    E6 = -0.021980907677118407

    LL = E0*B + E2*sin(2*B) + E4*sin(4*B) + E6*sin(6*B)
    x0 = mH*LL

    LhL = 3 + 6*floor(degrees(L)/6)
    Zone = 30 + (3+LhL)/6

    DL = L - radians(LhL)

    etaq = eq*cB**2
    Nq = c/sqrt(1+etaq)

    x2 = mH/2*Nq*tB*cB**2
    x4 = mH/24*Nq*tB*cB**4*(5-tB**2+9*etaq)
    x6 = mH/720*Nq*tB*cB**6*(61-58*tB**2+tB**4)

    N = x0 + x2*DL**2 + x4*DL**4 + x6*DL**6

    y1 = mH*Nq*cB
    y3 = mH/6*Nq*cB**3*(1-tB**2+etaq)
    y5 = mH/120*Nq*cB**5*(5-18*tB**2+tB**4+etaq*(14-58*tB**2))

    E = y1*DL + y3*DL**3 + y5*DL**5 + 500000

    return N, E, Zone


def unbing(latlon):
    "Repapiert Bing-Koordinaten"
    # Aus einer kaputten Bing-Maps-Koordinate wie beispielsweise
    # "51,468936, 7,447377" wird eine verwertbare: "51.468936, 7.447377".
    if latlon.count(",") == 3:
        a, b, c, d = latlon.split(",")
        latlon = a+"."+b+", "+c+"."+d
    # Die Website espg.io lässt das Trennzeichen ganz weg und verwendet eine
    # umgedrehte Reihenfolge: "7.268823 51.447735".
    elif latlon.count(",") == 0: # espg.io
        latlon = ", ".join(latlon.split()[::-1])
    return latlon

# Schritt 1: Wie soll das Kind heißen?
print("\nGeben Sie einen Basisnamen für die erzeugten Dateien an!\n"
      "Existierende Dateien mit diesem Namen und den Endungen .dxf, \n"
      ".xyz, .png, .log, .txt und .stl werden ohne Warnung überschrieben.\n")

basisname = input("Dateiname ohne Endung: ")
log("Basisname: "+basisname, sichtbar=False)

# Ordner zum Zwischenspeichern der xyz-Daten
ordner = "DGM1"

if os.path.isdir(ordner):
    log(f"Verwende vorhandenen Ordner {ordner} für die DGM-Daten.")
else:
    try:
        os.makedirs(ordner)
        log(f"Ordner {ordner} wurde zur Zwischenspeicherung der DGM-Daten "
            "angelegt.")
    except FileExistsError:
        log(f"FEHLER: Kann Ordner {ordner} nicht anlegen. "
            "Entfernen Sie die gleichnamige Datei!")
        sys.exit()

# Schritt 3: Welches Gebiet wollen wir modellieren?
print("\nGeben Sie zwei gegenüberliegende Eckpunkte des zu modellierenden\n"
      "Rechtecks an! Das Eingabeformat ist (Breite, Länge) in Dezimalgrad,\n"
      "also beispielsweise 51.335757,7.479087 – Sie können die Koordinaten\n"
      "direkt aus Google Maps oder Ihrem GPS-Gerät übernehmen.\n")

try:
    lat1, lon1 = eval(unbing(input("Erstes Eckpunktkoordinatenpaar: ")))
    lat2, lon2 = eval(unbing(input("Gegenüberliegendes Koordinatenpaar: ")))
except:
    print("Es wurden keine zwei mit einem Komma getrennte Zahlen erkannt.")
    sysexit()
    
# Die Eingabe oben war beliebig, wir brauchen aber gleich die Südwestecke
# unten links und die Nordostecke oben rechts:
ul_lat, or_lat = sorted((lat1, lat2))
ul_lon, or_lon = sorted((lon1, lon2))
log(f"Geokoordinaten: {ul_lat},{ul_lon} {or_lat},{or_lon}", sichtbar=False)

# Umrechnung der Dezimalgradkoordinaten ins UTM-System; aus Längen- und
# Breitengrad werden ein Nordwert, ein Ostwert und eine Zonennummer.
ul_n, ul_e, zn = map(int, utm(ul_lat, ul_lon))

# Das gleiche noch einmal für die Nordostecke unseres Rechtecks.
or_n, or_e, zn = map(int, utm(or_lat, or_lon))

log("\nLage und Größe des ausgewählten Rechtecks:")
log(f"UTM-Koordinaten: {ul_e},{ul_n} {or_e},{or_n} in Zone {zn}")
log(f"Ausdehnung Ost-West: {or_e-ul_e} m")
log(f"Ausdehnung Nord-Süd: {or_n-ul_n} m")
log(f"Fläche: {(or_e-ul_e)*(or_n-ul_n)} m²")

# Schritt 4: Liste der zu öffnenden XYZ-Dateien zusammenbauen.
# Bei den NRW-Höhendaten kann man am Dateinamen erkennen, ob gesuchte
# Punkte in ihnen enthalten sein könnten.
# Der Dateiname gibt die untere linke Ecke einer 1000x1000-m²-Kachel an.
# Beispiel für einen Dateinamen: dgm1_32_368_5700_2_nw.xyz
# Hier sind die Ostwerte 368000 bis 368999 und die Nordwerte
# 5700000 bis 5700999 enthalten.

log("\nListe der benötigten Geländekacheln:", sichtbar=False)
xyz_Liste = []
e_min = ul_e // 1000
e_max = or_e // 1000
n_min = ul_n // 1000
n_max = or_n // 1000
for e in range(e_min, e_max+1):
    for n in range(n_min, n_max+1):
        s = f"dgm1_{zn}_{e}_{n}_1_nw.xyz.gz"
        xyz_Liste.append(s)
        log(s, sichtbar=False)

# Die digitalen Orthophotos enthalten 1 km² große Kacheln mit 10 cm Auflösung
dop_Liste = []
if gdal:
    log("\nListe der benötigten Orthophotos:", sichtbar=False)
    for e in range(e_min, e_max+1):
        for n in range(n_min, n_max+1):
            s = f"dop10rgbi_32_{e}_{n}_1_nw.jp2"
            dop_Liste.append(s)
            log(s, sichtbar=False)
            
    log("\nLage des Bildausschnitts in den Orthophotos:")
    clip_l = 10 * (ul_e - 1000 * e_min)
    clip_r = 10 * (1000 * e_max - or_e + 1000)
    clip_o = 10 * (1000 * n_max - or_n + 1000)
    clip_u = 10 * (ul_n - 1000 * n_min)
    clip_h = (or_n - ul_n) * 10
    clip_b = (or_e - ul_e) * 10
    log(f"Beschnitt links: {clip_l:6} Pixel")
    log(f"Beschnitt rechts:{clip_r:6} Pixel")
    log(f"Beschnitt oben:  {clip_o:6} Pixel")
    log(f"Beschnitt unten: {clip_u:6} Pixel")
    log(f"Bildbreite:{clip_b:12} Pixel")
    log(f"Bildhöhe:  {clip_h:12} Pixel")

# Sind alle Dateien vorhanden?
print("\nUntersuche Vollständigkeit der Downloads …")
# Liste fehlender Geländekacheln und Orthophotos
fehlende_xyz = []
fehlende_dop = []

# Für alle benötigten Kacheln:
for Typ, Liste, fehlende, url in (
    ("Geländekacheln",
     xyz_Liste, fehlende_xyz, "https://www.opengeodata.nrw.de/"
                              "produkte/geobasis/hm/dgm1_xyz/dgm1_xyz/"),
    ("Orthophotos",
     dop_Liste, fehlende_dop, "https://www.opengeodata.nrw.de/"
                              "produkte/geobasis/lusat/dop/dop_jp2_f10/")):
    for name in Liste:
        # Ist diese Kachel im angegebenen Ordner vorhanden?
        if not os.path.isfile(os.path.join(ordner,name)):
            # Nein? Name an Liste fehlender Kacheln anhängen
            fehlende.append(name)

    # Sind Dateien herunterzuladen?
    if fehlende:
        print(f"\n{Typ} werden heruntergeladen:\n")

        for i, name in enumerate(fehlende, start=1): 

            # Datei wird heruntergeladen.
            print(f"{name} ({i}/{len(fehlende)}) …")
            dateiname = os.path.join(ordner,name)
            try:
                with urllib.request.urlopen(f"{url}{name}") as response, \
                     open(dateiname, 'wb') as datei:
                    datei.write(response.read())
            except urllib.error.URLError:
                print("*** FEHLER ***\n")
                print("Der Downloadversuch ist gescheitert. Bitte laden Sie\n",
                      f"{url}{name}\n"
                      "mit einem Webbrowser herunter und verschieben Sie die\n"
                      "Datei in den Ordner\n",
                      f"{ordner}.\n")
                print("Das Programm versucht nun, die Datei von einem "
                      "Webbrowser\n"
                      "herunterladen zu lassen.\n"
                      "Starten Sie es nach dem Download erneut!\n")
                webbrowser.open(url)
                sysexit()
                
print("… alle benötigten Dateien sind vorhanden.")
        
# Schritt 5: Modelldetails festlegen
print("\nDie horizontale Auflösung der Höhendaten beträgt einen Meter.\n"
      "Die Auflösung der Orthophotos ist sogar zehnmal feiner. Das kann bei\n"
      "größeren Flächen zu extremen Dateigrößen und Verarbeitungszeiten der\n"
      "erzeugten Modelle führen. Tipp: beginnen Sie mit ungefähr 1000\n"
      "Höhenpunkten auf der gesamten Fläche und erhöhen Sie die Auflösung in\n"
      "weiteren Durchläufen schrittweise.\n")

# Vorschlag für die horizontale Auflösung ermitteln; es sollten etwa 1000
# Punkte in der gesamten Fläche liegen.
kl0 = max(1, int(((or_e - ul_e) * (or_n - ul_n) / 1000) ** 0.5))
p0 = ((or_e - ul_e) // kl0) * ((or_n - ul_n) // kl0)
print(f"Bei {kl0} m Auflösung würden Sie beispielsweise {p0} Höhenpunkte "
      "erhalten.\n")

try:
    kl = int(input("Geben Sie einen ganzzahligen Wert ein [m]: "))
except:
    kl = 1

print("\nDie vertikale Auflösung der Daten beträgt einen Zentimeter. Das ist\n"
      "normalerweise in Ordnung. Für einen Höhenschichteneffekt wie bei\n"
      "architektonischen Geländemodellen kann dieser Wert auch auf 100 oder\n"
      "mehr geändert werden, je nach gewünschter Effektstärke.\n"
      "Nebeneffekt: Das CAD-Volumenmodell benötigt bei größeren Werten\n"
      "deutlich weniger Speicherplatz.\n")

try:
    kh = int(input("Geben Sie einen ganzzahligen Wert ein [cm]: "))
except:
    kh = 1

# Höhe und Breite als Vielfaches der gewählten Auflösung berechnen
xmax = or_e - (or_e - ul_e) % kl
ymax = or_n - (or_n - ul_n) % kl

log(f"Horizontale Auflösung [m]: {kl}")
log(f"Vertikale Auflösung [cm]: {kh}")
log(f"\nAusdehnung Ost-West: {xmax-ul_e} m")
log(f"Ausdehnung Nord-Süd: {ymax-ul_n} m")
fqm = (xmax - ul_e) * (ymax - ul_n)
log(f"Fläche: {fqm} m² bzw. {fqm/1e6:.3f} km²")

# Schritt 6: Verarbeitung der Eingabedateien.
# Alle gefundenen Höhenwerte werden zunächst in ein Dictionary
# geschrieben, aus dem sie für die einzelnen Dateien wieder
# ausgelesen werden.

D = {}

# Die Höhe der Unterseite ist nicht null, sondern orientiert sich
# am tatsächlichen Gelände.

minh = float("inf")
maxh = -float("inf")

# Schleife über alle zu verwendenden Eingabedateien
for i, dateiname in enumerate(xyz_Liste, start=1):
    with gzip.open(os.path.join(ordner, dateiname)) as dgm:
        log(f"Verarbeite Kachel {dateiname} ({i}/{len(xyz_Liste)}) …")
        for zeile in dgm:
            # Beispiel für eine Zeile aus Bochum:
            # 372000.00 5706000.00   61.32
            try:
                x, y, h = zeile.split()
                x = int(float(x))
                y = int(float(y))
                h = float(h)
            except ValueError:
                print("Abbruch, falsches Format:",zeile)
                break
            # Koordinaten der Zeile im gesuchten Rechteck?
            # print(ul_e,x,or_e,"#",ul_n,y,or_n)
            if ul_e <= x <= or_e and ul_n <= y <= or_n:
                # Liegt der Punkt im ausgedünnten Horizontalraster?
                if not (x - ul_e) % kl and not (y - ul_n) % kl:
                    # Punkt wird bei gewählter Auflösung verwendet
                    if kh != 1:
                        # Höhenwerte werden gerundet
                        h = round(round(h*100/kh)*kh/100, 2)
                    D[x,y] = h
                    minh = min(h, minh)
                    maxh = max(h, maxh)

log(f"Größte gefundene Höhe: {maxh:.2f} Meter")
log(f"Kleinste gefundene Höhe: {minh:.2f} Meter")

# "echten" minh-Wert für später merken
minhs = minh
# Unterkante des Modells absenken:
minh = 10 * floor(minh / 10) - 10

log(f"Setze Unterkante auf {minh:.2f} Meter.")
log(f"Neue Modellhöhe: {maxh-minh:.2f} Meter")

##############################################################
# Die Verarbeitung der Eingabedateien ist nun abgeschlossen. #
##############################################################

# Schritt 7: Erzeugen der Ausgabedateien.

##########################
# Matplotlib-Höhendiagramm

# Wenn der Matplotlib-Import erfolgreich war:
if plt:
    # Diagramm anzeigen
    print("\nHöhendiagramm wird erzeugt.")
    # Keine Interaktion, Diagramm nur anzeigen …
    plt.rcParams['toolbar'] = 'None'
    # Einheitlicher Maßstab für x und y
    plt.axis('equal')

    # Für das Diagramm wird die untere linke Ecke auf (0,0)
    # gesetzt. Wer kann schon was mit UTM-Koordinaten anfangen?
    
    # Liste der x-Werte
    xi = list(range(0, or_e - ul_e + 1, kl))
    # Liste der y-Werte
    yi = list(range(0, or_n - ul_n + 1, kl))
    # Matrix der Höhenwerte für alle x-y-Paare
    zi = [[D[(x + ul_e, y + ul_n)] for x in xi] for y in yi]

    # Anzahl der Höhenlinien: etwa 10 (7 bis 14)
    nh = (maxh - minhs) * 100
    while nh > 70: nh = int(nh / 10)
    if nh > 35: nh = int(nh / 5)
    if nh > 14: nh = int(nh / 2)
    # print("Zeichne",nh,"Höhenlinien.")

    # Höhenlinien
    plt.contour(xi, yi, zi, nh, linewidths=1, colors="k")
    # farbige Oberfläche
    plt.pcolormesh(xi, yi, zi, cmap = plt.get_cmap('terrain'),
                   shading="auto")
    # Legende
    plt.colorbar()
    # Überschrift
    plt.title(f"Ausschnittgröße {xmax-ul_e}×{ymax-ul_n} m\n"
              f"(0,0) bei {ul_lat}° Nord, {ul_lon}° Ost")
    plt.gcf().canvas.set_window_title(f'Höhendiagramm {basisname}')
    # 2020-06-16 Die folgende Kontrollanzeige entfernt. Irritiert nur.
    # Das Höhendiagramm anzeigen und zurück zum Programm …
    # plt.show(block=False)
    # Bitte nicht erst auf eine günstige Gelegenheit zur Anzeige warten,
    # sondern das Diagramm sofort anzeigen!
    # plt.gcf().canvas.flush_events()
    # Bild als PNG speichern
    ausname = basisname+".png"
    log(f"Schreibe Höhendiagramm: {ausname}")
    plt.savefig(ausname, bbox_inches='tight', dpi=600)
else:
    print("\nEin Höhendiagramm kann nicht erzeugt werden,\n"
          " weil Matplotlib nicht installiert ist.\n")

    
###########
# XYZ-Datei

# Alle verwendeten Punkte als simple XYZ-Datei sichern

ausname = basisname+".xyz"
log(f"Schreibe XYZ-Ausgabedatei: {ausname}")

with open(ausname,"w") as aus:
    for x in range(ul_e, or_e+1, kl):
        for y in range(ul_n, or_n+1, kl):
            aus.write(f"{x} {y} {D[(x,y)]}\n")

###############
# BIM-TXT-Datei            

# XYZ-Daten für BricsCAD mit Komma trennen und zum Koordinatenursprung
# verschieben.

ausname = basisname+".txt"
log(f"Schreibe BricsCAD-BIM-Geländedatei: {ausname}")

with open(ausname,"w") as aus:
    for x in range(ul_e, or_e+1, kl):
        for y in range(ul_n, or_n+1, kl):
            aus.write(f"{x-ul_e},{y-ul_n},{D[(x,y)]}\n")

############
# DXF-Export

ausname = basisname+".dxf"
log(f"Schreibe DXF-Datei mit 3D-Flächen: {ausname}")

# Weil 3D-Solids in einem obskuren „Geheimformat“ gespeichert werden, wird
# hier nur die umhüllende Fläche erzeugt. Diese muss im CAD-Programm mittels
# der Befehle REGION ALLE und DMHEFTEN bzw. FLÄCHEFORM umgewandelt werden,
# um einen richtigen Geländekörper zu erhalten.

def hf(h):
    # "Höhenfarbe"
    # Gibt abhängig von der relativen Höhe eine von 17 Indexfarben zurück,
    # Erscheinungsbild ähnelt dem topographischer Karten.
    return (194, 182, 170, 160, 150, 134, 112, 100, 91, 71, 61, 50, 40, 42,
            31, 9, 254, 254)[int(17*(h-minhs)/(maxh-minhs))]

with open(ausname,"w") as aus:
    # Kopf der DXF-Datei schreiben
    aus.write("0\nSECTION\n2\nHEADER\n"
              "9\n$ACADVER\n1\nAC1006\n"
              "9\n$INSBASE\n10\n0.0\n20\n0.0\n30\n0.0\n"
              "9\n$INSUNITS\n70\n6\n"
              "9\n$EXTMIN\n10\n0.0\n20\n0.0\n"
              f"9\n$EXTMAX\n10\n{xmax-ul_e}\n20\n{ymax-ul_n}\n"
              "9\n$LIMMIN\n10\n0.0\n20\n0.0\n"
              f"9\n$LIMMAX\n10\n{xmax-ul_e}\n20\n{ymax-ul_n}\n"
              "0\nENDSEC\n"
              "0\nSECTION\n2\nTABLES\n"
              "0\nTABLE\n2\nLAYER\n"
              "70\n2\n"
              "0\nLAYER\n2\nGelaendemodell.py\n"
              "70\n64\n"
              "62\n7\n"
              "6\nCONTINUOUS\n"
              "0\nENDTAB\n"
              "0\nTABLE\n2\nVPORT\n"
              "70\n1\n0\nVPORT\n2\n*ACTIVE\n"
              "70\n0\n10\n0.0\n20\n0.0\n11\n1.0\n21\n1.0\n"
              f"12\n{(xmax-ul_e)/2}\n" # Mitte x
              f"22\n{(ymax-ul_n)/2}\n" # Mitte y
              "13\n0.0\n23\n0.0\n14\n0.5\n24\n0.5\n15\n0.5\n25\n0.5\n"
              "16\n0.0\n26\n0.0\n36\n1.0\n17\n0.0\n27\n0.0\n37\n0.0\n"
              f"40\n{(ymax-ul_n)*1.01}\n" # Höhe des Bereichs
              f"41\n{(xmax-ul_e)/(ymax-ul_n)}\n" # Breitenverhältnis
              "42\n50.0\n43\n0.0\n44\n0.0\n50\n0.0\n51\n0.0\n"
              "71\n0\n72\n100\n73\n1\n74\n3\n75\n0\n76\n0\n77\n0\n78\n0\n"
              "0\nENDTAB\n"
              "0\nENDSEC\n"
              "0\nSECTION\n2\nENTITIES\n")
    
    # Geländeoberfläche
    for x in range(ul_e,xmax+1-kl,kl):
      xu = x - ul_e
      for y in range(ul_n,ymax+1-kl,kl):
        yu = y - ul_n
        h1 = D[x,y]
        h2 = D[x+kl,y]
        h3 = D[x,y+kl]
        h4 = D[x+kl,y+kl]
        # Die vier Eckpunkte werden auf zwei Dreiecke aufgeteilt.
        # Als Diagonale wird die kürzere der beiden Verbindungen
        # 1-4 bzw. 2-3 gewählt:
        # entweder
        #     3 4         4     3 4
        #     1 2   =   1 2  +  1
        # oder
        #     3 4       3       3 4
        #     1 2   =   1 2  +    2
        # Ein Dreieck ist im DXF-Format ein Viereck, bei dem der letzte Punkt
        # doppelt vorkommt.
        if abs(h4-h1) < abs(h2-h3):
            # Diagonale 1-4
            aus.write("0\n3DFACE\n8\nGelaendemodell.py\n"
                      f"62\n{hf((h1+h4+h2)/3)}\n"
                      f"10\n{xu}\n20\n{yu}\n30\n{h1}\n"
                      f"11\n{xu+kl}\n21\n{yu+kl}\n31\n{h4}\n"
                      f"12\n{xu+kl}\n22\n{yu}\n32\n{h2}\n"
                      f"13\n{xu+kl}\n23\n{yu}\n33\n{h2}\n")
            aus.write("0\n3DFACE\n8\nGelaendemodell.py\n"
                      f"62\n{hf((h1+h3+h4)/3)}\n"
                      f"10\n{xu}\n20\n{yu}\n30\n{h1}\n"
                      f"11\n{xu}\n21\n{yu+kl}\n31\n{h3}\n"
                      f"12\n{xu+kl}\n22\n{yu+kl}\n32\n{h4}\n"
                      f"13\n{xu+kl}\n23\n{yu+kl}\n33\n{h4}\n")
        else:
            # Diagonale 2-3
            aus.write("0\n3DFACE\n8\nGelaendemodell.py\n"
                      f"62\n{hf((h1+h3+h2)/3)}\n"
                      f"10\n{xu}\n20\n{yu}\n30\n{h1}\n"
                      f"11\n{xu}\n21\n{yu+kl}\n31\n{h3}\n"
                      f"12\n{xu+kl}\n22\n{yu}\n32\n{h2}\n"
                      f"13\n{xu+kl}\n23\n{yu}\n33\n{h2}\n")
            aus.write("0\n3DFACE\n8\nGelaendemodell.py\n"
                      f"62\n{hf((h2+h3+h4)/3)}\n"
                      f"10\n{xu+kl}\n20\n{yu}\n30\n{h2}\n"
                      f"11\n{xu}\n21\n{yu+kl}\n31\n{h3}\n"
                      f"12\n{xu+kl}\n22\n{yu+kl}\n32\n{h4}\n"
                      f"13\n{xu+kl}\n23\n{yu+kl}\n33\n{h4}\n")

      
    # Unterseite
    aus.write("0\n3DFACE\n8\nGelaendemodell.py\n"
              "62\n47\n"
              f"10\n0\n20\n0\n30\n{minh}\n"
              f"11\n{xmax-ul_e}\n21\n0\n31\n{minh}\n"
              f"12\n{xmax-ul_e}\n22\n{ymax-ul_n}\n32\n{minh}\n"
              f"13\n0\n23\n{ymax-ul_n}\n33\n{minh}\n")
        
    # Linke Wand
    for y in range(ul_n,ymax+1-kl,kl):
        yu = y-ul_n
        h1 = D[ul_e,y]
        h2 = D[ul_e,y+kl]
        aus.write("0\n3DFACE\n8\nGelaendemodell.py\n"
                  "62\n47\n"
                  f"10\n0\n20\n{yu}\n30\n{minh}\n"
                  f"11\n0\n21\n{yu+kl}\n31\n{minh}\n"
                  f"12\n0\n22\n{yu+kl}\n32\n{h2}\n"
                  f"13\n0\n23\n{yu}\n33\n{h1}\n")
        
    # Rechte Wand
    for y in range(ul_n,ymax+1-kl,kl):
        xu = xmax-ul_e
        yu = y-ul_n
        h1 = D[xmax,y]
        h2 = D[xmax,y+kl]
        aus.write("0\n3DFACE\n8\nGelaendemodell.py\n"
                  "62\n47\n"
                  f"10\n{xu}\n20\n{yu}\n30\n{minh}\n"
                  f"11\n{xu}\n21\n{yu}\n31\n{h1}\n"
                  f"12\n{xu}\n22\n{yu+kl}\n32\n{h2}\n"
                  f"13\n{xu}\n23\n{yu+kl}\n33\n{minh}\n")
        
    # Vordere Wand
    for x in range(ul_e,xmax+1-kl,kl):
        h1 = D[x,ul_n]
        h2 = D[x+kl,ul_n]
        xu = x - ul_e
        aus.write("0\n3DFACE\n8\nGelaendemodell.py\n"
                  "62\n47\n"
                  f"10\n{xu}\n20\n0\n30\n{minh}\n"
                  f"11\n{xu}\n21\n0\n31\n{h1}\n"
                  f"12\n{xu+kl}\n22\n0\n32\n{h2}\n"
                  f"13\n{xu+kl}\n23\n0\n33\n{minh}\n")
        
    # Hintere Wand
    for x in range(ul_e,xmax+1-kl,kl):
        xu = x - ul_e
        yu = ymax - ul_n
        h1 = D[x,ymax]
        h2 = D[x+kl,ymax]
        aus.write("0\n3DFACE\n8\nGelaendemodell.py\n"
                  "62\n47\n"
                  f"10\n{xu}\n20\n{yu}\n30\n{minh}\n"
                  f"11\n{xu+kl}\n21\n{yu}\n31\n{minh}\n"
                  f"12\n{xu+kl}\n22\n{yu}\n32\n{h2}\n"
                  f"13\n{xu}\n23\n{yu}\n33\n{h1}\n")
        
    aus.write("0\nENDSEC\n0\nEOF\n")
            

##################################
# Stereolitographiedatei STL ASCII   

ausname = basisname+".ascii.stl"
log(f"Schreibe STL-Datei für 3D-Druck (ASCII): {ausname}")

# Die Flächen umhüllen einen Quader, der unten auf Höhe minh aufliegt
# und oben durch die Geländeoberfläche abgeschnitten wird.

# Für die „Flächennormalen“ wird hier immer eine der Achsenrichtungen
# eingesetzt. Für die Geländeoberfläche ist das beispielsweise die
# nach oben gerichtete z-Achse (0, 0, 1).
# Der genaue Winkel ist auch gar nicht relevant. Hauptsache,
# die „Normale“ zeigt irgendwie aus dem umhüllten Körper heraus
# und nicht in ihn hinein.

with open(ausname,"w") as aus:
    aus.write("solid "+basisname+"\n")
    
    # Geländeoberfläche
    for x in range(ul_e,xmax+1-kl,kl):
      xu = x - ul_e
      for y in range(ul_n,ymax+1-kl,kl):
        yu = y - ul_n
        h1 = D[x,y]
        h2 = D[x+kl,y]
        h3 = D[x,y+kl]
        h4 = D[x+kl,y+kl]
        # Kürzere Diagonale wählen
        if abs(h4-h1) < abs(h2-h3):
            aus.write("facet normal 0 0 1\n"
                      "outer loop\n"
                      f"vertex {xu} {yu} {h1}\n"
                      f"vertex {xu+kl} {yu+kl} {h4}\n"
                      f"vertex {xu+kl} {yu} {h2}\n"
                      "endloop\n"
                      "endfacet\n")
            aus.write("facet normal 0 0 1\n"
                      "outer loop\n"
                      f"vertex {xu} {yu} {h1}\n"
                      f"vertex {xu} {yu+kl} {h3}\n"
                      f"vertex {xu+kl} {yu+kl} {h4}\n"
                      "endloop\n"
                      "endfacet\n")
        else:
            aus.write("facet normal 0 0 1\n"
                      "outer loop\n"
                      f"vertex {xu} {yu} {h1}\n"
                      f"vertex {xu} {yu+kl} {h3}\n"
                      f"vertex {xu+kl} {yu} {h2}\n"
                      "endloop\n"
                      "endfacet\n")
            aus.write("facet normal 0 0 1\n"
                      "outer loop\n"
                      f"vertex {xu+kl} {yu} {h2}\n"
                      f"vertex {xu} {yu+kl} {h3}\n"
                      f"vertex {xu+kl} {yu+kl} {h4}\n"
                      "endloop\n"
                      "endfacet\n")
            
        # Was für ein Aufwand für zwei popelige Dreiecke!

    # Die Unterseite und die Seitenflächen wiederholen im Moment
    # die Unterteilung der Oberseite. Dadurch hat die STL-Datei fast
    # doppelt so viele Dreiecke wie eigentlich nur nötig wären.
    # Durch geschickte Aufteilung könnte die Unterseite mit nur zwei
    # Dreiecken realisiert werden, wobei die Seitenflächen mit rund der
    # Hälfte der Dreiecke auskommen könnten. Bei großen Geländesteigungen
    # gibt es da aber ein paar Herausforderungen an den Algorithmus.
        
    # Unterseite
    for x in range(ul_e,xmax+1-kl,kl):
      xu = x - ul_e
      for y in range(ul_n,ymax+1-kl,kl):
        yu = y - ul_n
        aus.write("facet normal 0 0 -1\n"
                  "outer loop\n"
                  f"vertex {xu} {yu} {minh}\n"
                  f"vertex {xu+kl} {yu} {minh}\n"
                  f"vertex {xu+kl} {yu+kl} {minh}\n"
                  "endloop\n"
                  "endfacet\n")
        aus.write("facet normal 0 0 -1\n"
                  "outer loop\n"
                  f"vertex {xu} {yu} {minh}\n"
                  f"vertex {xu+kl} {yu+kl} {minh}\n"
                  f"vertex {xu} {yu+kl} {minh}\n"
                  "endloop\n"
                  "endfacet\n")
        
    # Linke Wand
    for y in range(ul_n,ymax+1-kl,kl):
        yu = y-ul_n
        h1 = D[ul_e,y]
        h2 = D[ul_e,y+kl]
        aus.write("facet normal -1 0 0\n"
                  "outer loop\n"
                  f"vertex 0 {yu} {minh}\n"
                  f"vertex 0 {yu+kl} {h2}\n"
                  f"vertex 0 {yu} {h1}\n"
                  "endloop\n"
                  "endfacet\n")
        aus.write("facet normal -1 0 0\n"
                  "outer loop\n"
                  f"vertex 0 {yu} {minh}\n"
                  f"vertex 0 {yu+kl} {minh}\n"
                  f"vertex 0 {yu+kl} {h2}\n"
                  "endloop\n"
                  "endfacet\n")
        
    # Rechte Wand
    for y in range(ul_n,ymax+1-kl,kl):
        xu = xmax-ul_e
        yu = y-ul_n
        h1 = D[xmax,y]
        h2 = D[xmax,y+kl]
        aus.write("facet normal 1 0 0\n"
                  "outer loop\n"
                  f"vertex {xu} {yu} {minh}\n"
                  f"vertex {xu} {yu} {h1}\n"
                  f"vertex {xu} {yu+kl} {h2}\n"
                  "endloop\n"
                  "endfacet\n")
        aus.write("facet normal 1 0 0\n"
                  "outer loop\n"
                  f"vertex {xu} {yu} {minh}\n"
                  f"vertex {xu} {yu+kl} {h2}\n"
                  f"vertex {xu} {yu+kl} {minh}\n"
                  "endloop\n"
                  "endfacet\n")
        
    # Vordere Wand
    for x in range(ul_e,xmax+1-kl,kl):
        h1 = D[x,ul_n]
        h2 = D[x+kl,ul_n]
        xu = x - ul_e
        aus.write("facet normal 0 -1 0\n"
                  "outer loop\n"
                  f"vertex {xu} 0 {minh}\n"
                  f"vertex {xu} 0 {h1}\n"
                  f"vertex {xu+kl} 0 {h2}\n"
                  "endloop\n"
                  "endfacet\n")
        aus.write("facet normal 0 -1 0\n"
                  "outer loop\n"
                  f"vertex {xu} 0 {minh}\n"
                  f"vertex {xu+kl} 0 {h2}\n"
                  f"vertex {xu+kl} 0 {minh}\n"
                  "endloop\n"
                  "endfacet\n")
        
    # Hintere Wand
    for x in range(ul_e,xmax+1-kl,kl):
        xu = x - ul_e
        yu = ymax - ul_n
        h1 = D[x,ymax]
        h2 = D[x+kl,ymax]
        aus.write("facet normal 0 1 0\n"
                  "outer loop\n"
                  f"vertex {xu} {yu} {minh}\n"
                  f"vertex {xu+kl} {yu} {h2}\n"
                  f"vertex {xu} {yu} {h1}\n"
                  "endloop\n"
                  "endfacet\n")
        aus.write("facet normal 0 1 0\n"
                  "outer loop\n"
                  f"vertex {xu} {yu} {minh}\n"
                  f"vertex {xu+kl} {yu} {minh}\n"
                  f"vertex {xu+kl} {yu} {h2}\n"
                  "endloop\n"
                  "endfacet\n")
    aus.write("endsolid "+basisname+"\n")

##################################
# Stereolitographiedatei STL Binär

# Hier geschieht das gleiche wie bei der STL-ASCII-Ausgabe, nur die
# erzeugte Datei ist kleiner.
#
# https://de.wikipedia.org/wiki/STL-Schnittstelle
#
# Die struct-Codes unten bedeuten:
#   "<"   niedrigwertiges Byte voran ("little-endian")
#   "I"   vorzeichenlose Ganzzahl (4 Byte)
#   "12f" 12 Gleitkommawerte (je 4 Byte)
#   "h"   kurze Ganzzahl (2 Byte)

ausname = basisname+".binär.stl"
log(f"Schreibe STL-Datei für 3D-Druck (binär): {ausname}")

with open(ausname,"wb") as aus:
    # 80 Bytes ungenutzter Header
    aus.write(b'\0' * 80)
    
    # Wie viele Dreiecke hat das Modell insgesamt?
    nx = (xmax - ul_e) // kl 
    ny = (ymax - ul_n) // kl 
    ngesamt = 4 * nx*ny + 4 * nx + 4 * ny
    aus.write(struct.pack('<I', ngesamt))
    
    # Geländeoberfläche
    for x in range(ul_e,xmax+1-kl,kl):
      xu = x - ul_e
      for y in range(ul_n,ymax+1-kl,kl):
        yu = y - ul_n
        h1 = D[x,y]
        h2 = D[x+kl,y]
        h3 = D[x,y+kl]
        h4 = D[x+kl,y+kl]
        if abs(h4-h1) < abs(h2-h3):
            aus.write(struct.pack("<12fh",
                                  0, 0, 1,
                                  xu, yu, h1,
                                  xu+kl, yu+kl, h4,
                                  xu+kl, yu, h2,
                                  0))
            aus.write(struct.pack("<12fh",
                                  0, 0, 1,
                                  xu, yu, h1,
                                  xu, yu+kl, h3,
                                  xu+kl, yu+kl, h4,
                                  0))
        else:
            aus.write(struct.pack("<12fh",
                                  0, 0, 1,
                                  xu, yu, h1,
                                  xu, yu+kl, h3,
                                  xu+kl, yu, h2,
                                  0))
            aus.write(struct.pack("<12fh",
                                  0, 0, 1,
                                  xu+kl, yu, h2,
                                  xu, yu+kl, h3,
                                  xu+kl, yu+kl, h4,
                                  0))
            
        
    # Unterseite
    for x in range(ul_e,xmax+1-kl,kl):
      xu = x - ul_e
      for y in range(ul_n,ymax+1-kl,kl):
        yu = y - ul_n
        aus.write(struct.pack("<12fh",
                              0, 0, -1,
                              xu, yu, minh,
                              xu+kl, yu+kl, minh,
                              xu+kl, yu, minh,
                              0))
        aus.write(struct.pack("<12fh",
                              0, 0, -1,
                              xu, yu, minh,
                              xu, yu+kl, minh,
                              xu+kl, yu+kl, minh,
                              0))
        
    # Linke Wand
    for y in range(ul_n,ymax+1-kl,kl):
        yu = y-ul_n
        h1 = D[ul_e,y]
        h2 = D[ul_e,y+kl]
        aus.write(struct.pack("<12fh",
                              -1, 0, 0, 
                              0, yu, minh,
                              0, yu+kl, h2,
                              0, yu, h1,
                              0))
        aus.write(struct.pack("<12fh",
                              -1, 0, 0,
                              0, yu, minh,
                              0, yu+kl, minh,
                              0, yu+kl, h2,
                              0))
        
    # Rechte Wand
    xu = xmax-ul_e
    for y in range(ul_n,ymax+1-kl,kl):
        yu = y-ul_n
        h1 = D[xmax,y]
        h2 = D[xmax,y+kl]
        aus.write(struct.pack("<12fh",
                              1, 0, 0, 
                              xu, yu, minh,
                              xu, yu, h1,
                              xu, yu+kl, h2,
                              0))
        aus.write(struct.pack("<12fh",
                              1, 0, 0,
                              xu, yu, minh,
                              xu, yu+kl, h2,
                              xu, yu+kl, minh,
                              0))
        
    # Vordere Wand
    for x in range(ul_e,xmax+1-kl,kl):
        h1 = D[x,ul_n]
        h2 = D[x+kl,ul_n]
        xu = x - ul_e
        aus.write(struct.pack("<12fh",
                              0, -1, 0, 
                              xu, 0, minh,
                              xu, 0, h1,
                              xu+kl, 0, h2,
                              0))
        aus.write(struct.pack("<12fh",
                              0, -1, 0,
                              xu, 0, minh,
                              xu+kl, 0, h2,
                              xu+kl, 0, minh,
                              0))
        
    # Hintere Wand
    yu = ymax - ul_n
    for x in range(ul_e,xmax+1-kl,kl):
        xu = x - ul_e
        h1 = D[x,ymax]
        h2 = D[x+kl,ymax]
        aus.write(struct.pack("<12fh",
                              0, 1, 0, 
                              xu, yu, minh,
                              xu+kl, yu, h2,
                              xu, yu, h1,
                              0))
        aus.write(struct.pack("<12fh",
                              0, 1, 0,
                              xu, yu, minh,
                              xu+kl, yu, minh,
                              xu+kl, yu, h2,
                              0))

##############################################
# Orthophotos

# Die Orthophotos können in BricsCAD und AutoCAD
# als Materialtextur der Geländeoberfläche verwendet werden
# sodass ein sehr realistischer Geländeeindruck entsteht.

# Kaum zu glauben, wie simpel das ist:

if gdal:
    try:
        dop_lange_Liste = [os.path.join(ordner, dop) for dop in dop_Liste]
        gdal.BuildVRT(basisname+".vrt", dop_lange_Liste)
        log(f"Schreibe beschnittenes kombiniertes Orthophoto: {basisname}.jpg")
        gdal.Translate(basisname+".jpg",
                       basisname+".vrt",
                       # Lage im Ursprungsmosaik
                       srcWin = (clip_l, clip_o,
                                 clip_b, clip_h),
                       # automatische Helligkeitsspreizung
                       scaleParams = [[]],
                       # Reduzierung der Ausgabegröße durch Begrenzung der 
                       # Bildauflösung auf 10 Pixel pro Messpunktabstand
                       width = min(clip_b, int(10 * (or_e - ul_e) / kl)),
                       height = 0
                       )
    except Exception as e:
        log(f"Beim Zusammenstellen des Orthophotos gab es einen Fehler: {e}")         
else:
    log("Das Python-Modul 'osgeo.gdal' kann nicht geladen werden.\n"
        "Die J2P-Orthophotos werden daher nicht zu einer jpg-Geländetextur"
        " verarbeitet.\n"
        "Abhilfe für MS Windows:   pip install gdal\n"
        "Abhilfe für Ubuntu Linux: apt install python3-gdal.")


log("Programmlauf erfolgreich beendet.\n\n")
print("Die Ausgabedateien können nun weiterverarbeitet werden.")

input("\nProgramm schließen mit [Enter]")
